(* Weimer's Elkhound Ocaml token.list -> hideous function generator *)

open Lexer         (* tToken, readToken *)
open Lexerint      (* tLexerInterface *)
open Parsetables   (* tParseTables *)
open Useract       (* tUserActions *)
open Glr           (* tGLR, makeGLR, glrParse *)

open Ptreeact      (* tParseTreeLexer, makeParseTreeActions *)
open Ptreenode     (* tPTreeNode, printTree *)

open Otype


class tLexer lexbuf =
object (self)
  inherit tLexerInterface

  val lexbuf:Lexing.lexbuf = lexbuf 

  method getToken() : unit =
  begin
    let t:tToken = (readToken lexbuf) in
    tokType <- (tokenKind t);
    match t with
	| BIG_IDENT(o) -> sval <- (Obj.repr o)
	| SMALL_IDENT(o) -> sval <- (Obj.repr o)

 | _ -> ()
 
  end

  method tokenDesc() : string =
  begin
    let kindDesc:string = (self#tokenKindDesc tokType) in
    kindDesc
  end

  method tokenKindDesc (kind:int) : string =
  begin
    (Lexer.tokenKindDesc kind)
  end
end

let split name =
  try 
    let base = Filename.chop_extension name in
    let ext = String.sub name ((String.length base)+1)
      ((String.length name) - ((String.length base)+1))
    in 
    base,ext
  with _ -> name,""

let rec te_to_string te = match te with
  | Tick(s) -> Printf.sprintf "'%s" s
  | Underscore -> "_"
  | FunType(a,b) -> Printf.sprintf "%s -> %s" (te_to_string a)
                        (te_to_string b)
  | Tuple([]) -> failwith "te_to_string empty tuple" 
  | Tuple(hd::tl) -> 
      Printf.sprintf "%s %s" (te_to_string hd)
        (List.fold_left (fun acc elt -> acc ^ " * " ^ (te_to_string elt))
          "" tl)
  | Constructor(s,[]) -> s
  | Constructor(s,lst) -> failwith "te_to_string constructor"

let main () = begin
  let files = ref [] in 
  let argDescr = [ ] in
  let give_help () = Arg.usage argDescr "" ; exit 1 in 
  let handleArg str = begin files := str :: !files end in
  Arg.parse argDescr handleArg "" ; 
  let (tables:tParseTables), (actions:tUserActions) =
    (Oparse.oparseParseTables, Oparse.oparseUserActions)
  in
  List.iter (fun filename -> 
    try 
      let inchan = open_in filename in 
      let f,e = split filename in 
      let lexbuf = Lexing.from_channel inchan in 
      let lex:tLexerInterface = ((new tLexer lexbuf) :> tLexerInterface) in
      (* prime the lexer: get first token *)
      (lex#getToken());
      (* get parse tables and user actions, depending on which
       * grammar the user wants to use *)
      (* substitute tree-building actions? *)
      let (lex:tLexerInterface), (actions:tUserActions) =
        (lex, actions)                           (* unchanged *)
      in 
      (* parse the input *)
      let (sval:tSemanticValue) = 
          (* use GLR *)
          let glr:tGLR = (makeGLR tables actions) in
          let treeTop: tSemanticValue ref = ref cNULL_SVAL in
          if (not (glrParse glr lex treeTop)) then (
            (failwith "GLR parse error")
          );
          !treeTop
      in
      close_in inchan ; 
      Printf.printf "Parsed.\n" ; flush stdout ; 
      (* interpret the result *)
      let find_shortcut s = 
        let shortcut = 
          if Hashtbl.mem Lexer.commentHT s then
            Hashtbl.find Lexer.commentHT s
          else 
            ""
        in
        let shortcut = 
          if shortcut = "" then String.lowercase s else shortcut
        in 
        shortcut
      in 

      let handle_it t cdl opt = begin
          Printf.printf "%s: %d tokens.\n" filename (List.length cdl) ;
          let i = ref 0 in 
          let out = open_out_bin (f ^ ".parser") in
          List.iter (fun (s,teo) -> 
            Printf.fprintf out " %3d : TOK_%-30s \"%s\" ;\n" !i s 
              (find_shortcut s) ; 
            incr i 
          ) cdl ; 
          (match opt with
            None -> 
              List.iter (fun (s,teo) -> match teo with
              | Some(te) -> 
                Printf.fprintf out "\ttoken(%s) TOK_%s;\n" 
                  (te_to_string te) s
              | None -> () 
              ) cdl ; 
          | Some(span) -> 
              List.iter (fun (s,teo) -> match teo with 
              | Some(te) -> 
                Printf.fprintf out "\ttoken(%s * %s) TOK_%s;\n" 
                  (te_to_string te) (te_to_string span) s
              | None -> 
                Printf.fprintf out "\ttoken(%s) TOK_%s;\n" 
                  (te_to_string span) s
              ) cdl ; 
          ) ;

          close_out out ;
          let out = open_out_bin (f ^ ".lexer") in 
          i := 0 ; 
          (if (opt = None) then 
            Printf.fprintf out "let tokenKind (t:%s) : int = match t with\n" t 
          else
            Printf.fprintf out "let tokenKind ((t,_):%s) : int = match t with\n" t );
          List.iter (fun (s,teo) ->
            Printf.fprintf out "| %s" s ;
            (if (teo = None) then ()
            else Printf.fprintf out " _") ;
            Printf.fprintf out " -> %d\n" !i ;
            incr i 
          ) cdl ; 

          Printf.fprintf out "let tokenKindDesc (t:int) : string = match t with\n" ;
          i := 0; 
          List.iter (fun (s,teo) ->
            Printf.fprintf out "| %d -> \"%s\"\n" !i (find_shortcut s);
            incr i 
          ) cdl ; 
          Printf.fprintf out "| _ -> failwith \"tokenKindDesc\"\n" ;

          (if (opt = None) then 
            Printf.fprintf out "let tokenDesc (t:%s) : string = match t with\n" t
          else
            Printf.fprintf out "let tokenDesc ((t,s):%s) : string = match t with\n" t );
          List.iter (fun (s,teo) ->
            match teo with
            | Some(Constructor("int",[])) -> 
              Printf.fprintf out "| %s(i) -> string_of_int i\n" s
            | Some(Constructor("Int32.t",[])) -> 
              Printf.fprintf out "| %s(i) -> Int32.to_string i\n" s
            | Some(Constructor("string",[])) -> 
              Printf.fprintf out "| %s(s) -> s\n" s
            | _ -> ()
          ) cdl ; 
          (if (opt = None) then 
            Printf.fprintf out "| _ -> tokenKindDesc (tokenKind t)\n" 
          else
            Printf.fprintf out "| _ -> tokenKindDesc (tokenKind (t,s))\n" );

          close_out out  ;

          let out = open_out_bin (f ^ ".lexint") in
          Printf.fprintf out "
class tLexer lexbuf =
object (self)
  inherit tLexerInterface

  val lexbuf:Lexing.lexbuf = lexbuf 

  method getToken() : unit =
  begin
    let t:%s = (readToken lexbuf) in
    tokType <- (tokenKind t);
    match t with\n" t ;
        (match opt with
        | None -> 
          List.iter (fun (s,teo) ->
            match teo with
            | Some(Constructor(x,[])) -> 
              Printf.fprintf out "\t| %s(o) -> sval <- (Obj.repr o)\n" s
            | _ -> ()
          ) cdl ; 
          Printf.fprintf out "\n | _ -> ()\n" 
        | Some(span) -> 
          List.iter (fun (s,teo) ->
            match teo with
            | Some(Constructor(x,[])) -> 
              Printf.fprintf out "\t| (%s(o)),span -> sval <- (Obj.repr (o,span))\n" s
            | _ -> ()
          ) cdl ; 
            Printf.fprintf out "\n | (_,span) -> sval <- (Obj.repr span)\n"  ;
        ) ;
          Printf.fprintf out " 
  end

  method tokenDesc() : string =
  begin
    let kindDesc:string = (self#tokenKindDesc tokType) in
    kindDesc
  end

  method tokenKindDesc (kind:int) : string =
  begin
    (Lexer.tokenKindDesc kind)
  end
end" ;
          close_out out ; 
      end in 

      begin
        let (tdll:Otype.type_def list list) = 
          ((Obj.obj sval) : Otype.type_def list list) in
        let tdl = List.flatten tdll in 
        match tdl with
        | [ (t      ,[],(OType(Tuple([tok;span])))) ;
            (t_union,[],(OUnion(cdl))) ] ->
            handle_it t cdl (Some(span))

        | [(t,[],(OUnion(cdl)))] -> 
          handle_it t cdl None

        | _ -> failwith "Input file does not contain a single union type"
      end 
    with e -> 
      Printf.printf "%s: %s\n" filename 
        (Printexc.to_string e) 
  ) !files

end ;;

main () ;; 
