(* lexer.mll *)
(* based on an example from camllex documentation *)

{
  (* this encodes both the token kind (the constructor name) and
   * its lexeme (the constructor arguments);
   * unlike lex, ocamllex does not separate the concepts
   *)

  (* The following functions should probably be implemented by
   * some automatic translator, given the definition of tToken.
   * ocamlyacc does this (to some degree), but Elkhound does not.
   *)
open Util 
include(token.type)

include(token.lexer)

  let commentContents = ref "" 

  let commentHT = Hashtbl.create 511 

}

let blank = [' ' '\012' ]
let letter = ['a'- 'z' 'A'-'Z']
let big_letter = ['A'-'Z' ]
let small_letter = ['a'-'z' ]
let letter = ( big_letter | small_letter )
let decdigit =  ['0'-'9'] 
let big_ident = (big_letter)(letter|decdigit|'_')* 
let small_ident = (small_letter)(letter|decdigit|'_')* 

rule readToken = parse
    blank               { adj (Lexing.lexeme lexbuf) () ;
                          readToken lexbuf }     (* skip whitespace *)
  | "(*"                { commentContents := "" ; 
                          adj "(*" () ; 
                          let _ = comment lexbuf in readToken lexbuf } 
  | '\t'                { tab (); readToken lexbuf }
  | '\n'
  | "\r\n"
  | '\r'                { newline () ; readToken lexbuf }
  | small_ident         { let res = match Lexing.lexeme lexbuf with
                          | "type" -> TYPE
                          | "and" -> AND
                          | "of" -> OF
                          | "mutable" -> MUTABLE
                          | id -> SMALL_IDENT(id)
                          in 
                          adj (Lexing.lexeme lexbuf) res 
                        }
  | big_ident           { let tok = (BIG_IDENT(Lexing.lexeme lexbuf)) in 
                          Hashtbl.add commentHT (Lexing.lexeme lexbuf) 
                            !commentContents ; 
                          commentContents := "" ; 
                          adj (Lexing.lexeme lexbuf) tok } 
  | "->"                { adj (Lexing.lexeme lexbuf) ARROW } 
  | ","                 { adj (Lexing.lexeme lexbuf) COMMA } 
  | "."                 { adj (Lexing.lexeme lexbuf) DOT } 
  | "|"                 { adj (Lexing.lexeme lexbuf) BAR} 
  | "'"                 { adj (Lexing.lexeme lexbuf) QUOTE }
  | "="                 { adj (Lexing.lexeme lexbuf) EQUALS }
  | "*"                 { adj (Lexing.lexeme lexbuf) STAR }
  | ";"                 { adj (Lexing.lexeme lexbuf) SEMICOLON }
  | ":"                 { adj (Lexing.lexeme lexbuf) COLON }
  | '('                 { adj (Lexing.lexeme lexbuf) LPAREN }
  | ')'                 { adj (Lexing.lexeme lexbuf) RPAREN }
  | '{'                 { adj (Lexing.lexeme lexbuf) LBRACE }
  | '}'                 { adj (Lexing.lexeme lexbuf) RBRACE }
  | eof                 { adj "" EOF }
  | _                   { lex_error (Printf.sprintf "invalid character '%s'"
  (Lexing.lexeme lexbuf)) } 
and comment = parse 	
      "*)"	{ adj "" () }
|     '\n'      { newline () ; comment lexbuf }
|     "(*"      { ignore (adj "" ()) ; 
                  let _ = comment lexbuf in comment lexbuf } 
|     eof       { lex_error "unterminated comment" } 
|     blank     { adj (Lexing.lexeme lexbuf) () ; comment lexbuf } 
|     _    	{ adj (Lexing.lexeme lexbuf) () ; 
                  commentContents := !commentContents ^ 
                    (Lexing.lexeme lexbuf ) ; comment lexbuf }
