open Pretty
open Otype

let rec d_type_def () (a1 : type_def) =
  let (a2, a3, a4) = a1 in
    text "(" ++ dprintf "\"%s\"" a2 ++ text "," ++
    dprintf "[@[%a@]]" (docList (chr ';' ++ break)
                       (fun elt -> dprintf "\"%s\"" elt)) a3 ++ text "," ++
    d_type_info () a4 ++ text ")"

and d_type_def_list () (a5 : type_def_list) =
  dprintf "[@[%a@]]" (docList (chr ';' ++ break) (fun elt -> d_type_def () elt))
                     a5

and d_type_info () (a6 : type_info) =
  d_type_rep () a6

and d_type_rep () (a7 : type_rep) =
  begin match a7 with
    OUnion a8 -> text "OUnion(" ++ 
                 dprintf "[@[%a@]]" (docList (chr ';' ++ break)
                                    (fun elt -> d_constr_decl () elt)) a8
                 ++ text ")"
  | ORecord a9 -> text "ORecord(" ++ 
                  dprintf "[@[%a@]]" (docList (chr ';' ++ break)
                                     (fun elt -> d_field_decl () elt)) a9
                  ++ text ")"
  | OType a10 -> text "OType(" ++  d_typexpr () a10 ++ text ")"
  end

and d_constr_decl () (a11 : constr_decl) =
  let (a12, a13) = a11 in
    text "(" ++ dprintf "\"%s\"" a12 ++ text "," ++
    dprintf "@[%a@]" (docOpt (fun () elt -> d_typexpr () elt)) a13 ++ text ")"

and d_field_decl () (a14 : field_decl) =
  let (a15, a16, a17) = a14 in
    text "(" ++ dprintf "%b" a15 ++ text "," ++ dprintf "\"%s\"" a16
    ++ text "," ++ d_polytypexpr () a17 ++ text ")"

and d_polytypexpr () (a18 : polytypexpr) =
  d_typexpr () a18

and d_typexpr () (a19 : typexpr) =
  begin match a19 with
    Tick a20 -> text "Tick(" ++  dprintf "\"%s\"" a20 ++ text ")"
  | Underscore -> text "Underscore"
  | FunType(a21,a22) -> text "FunType(" ++  d_typexpr () a21 ++ text "," ++
                        d_typexpr () a22 ++ text ")"
  | Tuple a23 -> text "Tuple(" ++ 
                 dprintf "[@[%a@]]" (docList (chr ';' ++ break)
                                    (fun elt -> d_typexpr () elt)) a23
                 ++ text ")"
  | Constructor(a24,a25) -> text "Constructor(" ++  dprintf "\"%s\"" a24
                            ++ text "," ++
                            dprintf "[@[%a@]]" (docList (chr ';' ++ break)
                                               (fun elt -> d_typexpr () elt))
                                               a25 ++ text ")"
  end

(** Different visiting actions. *)
type 'a visitAction = 
    SkipChildren                        (** Do not visit the children. Return 
                                            the node as it is. *)
  | DoChildren                          (** Continue with the children of this 
                                            node. Rebuild the node on return 
                                            if any of the children changes 
                                            (use == test) *)
  | ChangeTo of 'a                      (** Replace the expression with the 
                                            given one *)
  | ChangeDoChildrenPost of 'a * ('a -> 'a) (** First consider that the entire 
                                           exp is replaced by the first 
                                           parameter. Then continue with 
                                           the children. On return rebuild 
                                           the node if any of the children 
                                           has changed and then apply the 
                                           function on the node *)
class type otypeVisitor = object
  method v_type_def : (type_def) -> (type_def) visitAction
  method v_type_def_list : (type_def_list) -> (type_def_list) visitAction
  method v_type_info : (type_info) -> (type_info) visitAction
  method v_type_rep : (type_rep) -> (type_rep) visitAction
  method v_constr_decl : (constr_decl) -> (constr_decl) visitAction
  method v_field_decl : (field_decl) -> (field_decl) visitAction
  method v_polytypexpr : (polytypexpr) -> (polytypexpr) visitAction
  method v_typexpr : (typexpr) -> (typexpr) visitAction
end
class nopOtypeVisitor : otypeVisitor = object
  method v_type_def : ((type_def) -> ((type_def) visitAction)) =
  fun _ -> DoChildren
  method v_type_def_list : ((type_def_list) -> ((type_def_list) visitAction)) =
  fun _ -> DoChildren
  method v_type_info : ((type_info) -> ((type_info) visitAction)) =
  fun _ -> DoChildren
  method v_type_rep : ((type_rep) -> ((type_rep) visitAction)) =
  fun _ -> DoChildren
  method v_constr_decl : ((constr_decl) -> ((constr_decl) visitAction)) =
  fun _ -> DoChildren
  method v_field_decl : ((field_decl) -> ((field_decl) visitAction)) =
  fun _ -> DoChildren
  method v_polytypexpr : ((polytypexpr) -> ((polytypexpr) visitAction)) =
  fun _ -> DoChildren
  method v_typexpr : ((typexpr) -> ((typexpr) visitAction)) =
  fun _ -> DoChildren
end

(* visit all the nodes in a type expression *)
let doVisit (vis: 'b)
            (startvisit: 'a -> 'a visitAction)
            (children: 'b -> 'a -> 'a)
            (node: 'a) : 'a =
  let action = startvisit node in
  match action with
    SkipChildren -> node
  | ChangeTo node' -> node'
  | _ -> (* DoChildren and ChangeDoChildrenPost *)
      let nodepre = match action with
        ChangeDoChildrenPost (node', _) -> node'
      | _ -> node                                                                     in
      let nodepost = children vis nodepre in
      match action with
        ChangeDoChildrenPost (_, f) -> f nodepost
      | _ -> nodepost
let rec visit_type_def (vis : otypeVisitor) v =
  doVisit vis vis#v_type_def visit_children_type_def v
and visit_children_type_def (vis : otypeVisitor) v =
  let (a26, a27, a28) = v in
  let tmp_a26 = a26 (* no string visitor *) in
  let tmp_a27 = List.map (fun v -> v (* no string visitor *)) a27 in
  let tmp_a28 = visit_type_info vis a28 in
  if (tmp_a26 == a26 && tmp_a27 == a27 && tmp_a28 == a28) then v
  else (tmp_a26 , tmp_a27 , tmp_a28)
and visit_type_def_list (vis : otypeVisitor) v =
  doVisit vis vis#v_type_def_list visit_children_type_def_list v
and visit_children_type_def_list (vis : otypeVisitor) v =
  List.map (fun v -> visit_type_def vis v) v
and visit_type_info (vis : otypeVisitor) v =
  doVisit vis vis#v_type_info visit_children_type_info v
and visit_children_type_info (vis : otypeVisitor) v =
  visit_type_rep vis v
and visit_type_rep (vis : otypeVisitor) v =
  doVisit vis vis#v_type_rep visit_children_type_rep v
and visit_children_type_rep (vis : otypeVisitor) v =
  begin match v with
    OUnion a29 -> let tmp_a29 = List.map (fun v -> visit_constr_decl vis v) a29 in
                  if (tmp_a29 == a29) then v else (OUnion(tmp_a29))
  | ORecord a30 -> let tmp_a30 = List.map (fun v -> visit_field_decl vis v) a30 in
                   if (tmp_a30 == a30) then v else (ORecord(tmp_a30))
  | OType a31 -> let tmp_a31 = visit_typexpr vis a31 in
                 if (tmp_a31 == a31) then v else (OType(tmp_a31))
  end
and visit_constr_decl (vis : otypeVisitor) v =
  doVisit vis vis#v_constr_decl visit_children_constr_decl v
and visit_children_constr_decl (vis : otypeVisitor) v =
  let (a32, a33) = v in
  let tmp_a32 = a32 (* no string visitor *) in
  let tmp_a33 = match a33 with None -> a33
                 | Some(x) -> (let tmp_x = visit_typexpr vis x in
                if (tmp_x == x) then a33 else (Some((tmp_x)))) in
  if (tmp_a32 == a32 && tmp_a33 == a33) then v else (tmp_a32 , tmp_a33)
and visit_field_decl (vis : otypeVisitor) v =
  doVisit vis vis#v_field_decl visit_children_field_decl v
and visit_children_field_decl (vis : otypeVisitor) v =
  let (a34, a35, a36) = v in
  let tmp_a34 = a34 (* no bool visitor *) in
  let tmp_a35 = a35 (* no string visitor *) in
  let tmp_a36 = visit_polytypexpr vis a36 in
  if (tmp_a34 == a34 && tmp_a35 == a35 && tmp_a36 == a36) then v
  else (tmp_a34 , tmp_a35 , tmp_a36)
and visit_polytypexpr (vis : otypeVisitor) v =
  doVisit vis vis#v_polytypexpr visit_children_polytypexpr v
and visit_children_polytypexpr (vis : otypeVisitor) v =
  visit_typexpr vis v
and visit_typexpr (vis : otypeVisitor) v =
  doVisit vis vis#v_typexpr visit_children_typexpr v
and visit_children_typexpr (vis : otypeVisitor) v =
  begin match v with
    Tick a37 -> let tmp_a37 = a37 (* no string visitor *) in
                if (tmp_a37 == a37) then v else (Tick(tmp_a37))
  | Underscore -> v
  | FunType(a38,a39) -> let tmp_a38 = visit_typexpr vis a38 in
                        let tmp_a39 = visit_typexpr vis a39 in
                        if (tmp_a38 == a38 && tmp_a39 == a39) then v
                        else (FunType(tmp_a38 , tmp_a39))
  | Tuple a40 -> let tmp_a40 = List.map (fun v -> visit_typexpr vis v) a40 in
                 if (tmp_a40 == a40) then v else (Tuple(tmp_a40))
  | Constructor(a41,a42) -> let tmp_a41 = a41 (* no string visitor *) in
                            let tmp_a42 = List.map (fun v -> visit_typexpr vis v) a42 in
                            if (tmp_a41 == a41 && tmp_a42 == a42) then v
                            else (Constructor(tmp_a41 , tmp_a42))
  end
