open Otype
open Pretty

let index = ref 0 
let gensym () = begin 
  incr index ; 
  Printf.sprintf "a%d" !index
end 

let doc_from_ticks () arg = 
          (match arg with
           | [] -> nil
           | [one] -> dprintf "'%s@?" one 
           | many -> dprintf "(@[%a@])@?" 
              (docList (chr ',' ++ break) 
                (fun s -> dprintf "'%s" s)) many
          ) 


let doc_from_union vname cdl myfun = 
  dprintf "begin match %s with@!" vname ++ align ++ text "  " ++
  docList (line ++ text "| ") (fun (cname,cteo) -> match cteo with
    | None -> 
        dprintf "%s -> " cname 
        ++ align
        ++ (myfun cname [])
        ++ unalign
    | Some(Tuple(el)) -> 
        let en = List.map (fun e -> gensym()) el in 
        let name_type_list = List.combine en el in 
        dprintf "%s(%a) -> " cname 
          (docList (chr ',') (fun x -> text x)) en 
        ++ align
        ++ (myfun cname name_type_list)
        ++ unalign 
    | Some(te) -> 
        let tn = gensym () in 
        dprintf "%s %s -> " cname tn 
        ++ align
        ++ (myfun cname [(tn,te)])
        ++ unalign
  ) () cdl 
  ++ unalign ++ line ++ text "end" 

(* target printf *) 
let tprintf (fmt:string) (args:doc) = 
  if args = nil then
    dprintf "text \"%s\" " fmt
  else 
    (dprintf "dprintf \"%s\" " fmt) ++ align ++ args ++ unalign

(* emit a pretty-printer for a type_def list *)
let emit_pp (tdll : (type_def list) list) = 
  (docList (line) (fun tdl -> 
  begin
    let already_type_def = Hashtbl.create 511 in 
    let first = ref true in 

    let rec pp_td ((name,otl,ti) : type_def) = begin
      if Hashtbl.mem already_type_def name || name = "node" then 
        nil
      else begin
        Hashtbl.add already_type_def name true ; 
        let vname = gensym () in 
        (dprintf "%s d_%s @[%a() (%s : %a%s) =@]@!  " 
          (if !first then (first := false ; "let rec") else "and")
          name 
          (docList (nil) (fun s -> 
            dprintf "(@[d_%s@?:@?unit@?->@?'%s@?->@?doc@])@?" s s
          )) otl
          vname
          doc_from_ticks otl 
          name
          ) 
        ++ align 
        ++ pp_ti vname ti 
        ++ unalign 
        ++ line
      end 
    end and pp_ti (vname : string) (ti : type_info) = begin
      match ti with
      | OUnion(cdl) -> 
          doc_from_union vname cdl (fun cname n_t_lst -> match n_t_lst with
          | [] -> dprintf "text \"%s\"" cname
          | _ -> dprintf "text \"%s(\" ++ break ++ " cname
                ++ break 
                ++ (docList (break ++ text "++ text \",\" ++" ++ break)
                  (fun (nm,tp) -> pp_te nm tp) () n_t_lst)
                ++ break ++ text "++ text \")\"" 
          ) 
      | ORecord(fdl) ->  
          tprintf "{ " nil ++ text "++ align ++" ++ break ++
          (docList (break ++ text "++ text \";\"" ++ break ++ 
                    text "++ line ++" ++ line) 
            (fun (mut,fn,pte) ->
            text (Printf.sprintf "text \"%s = \"" fn) ++ 
            break ++ text "++ break ++" ++ break  ++
            pp_te (vname ^ "." ^ fn) pte) 
            () fdl)  ++
          break ++ text "++ unalign ++" ++ break ++ tprintf "; }"  nil 
      | OType(te) -> pp_te vname te 
    end and pp_te (vname : string) (te : typexpr) = begin
      match te with
      | Tick(s) -> 
            dprintf "(d_%s () %s)" s vname
      | Underscore -> tprintf "_" nil 
      | FunType(a,b) -> tprintf "fun" nil 
      | Tuple(el) -> 
          let en = List.map (fun e -> gensym()) el in 
          let name_type_list = List.combine en el in 
          dprintf "let (%a) = %s in@!  " 
            (docList (text "," ++ break) 
              (fun (nm,tp) -> text nm)) name_type_list
              vname 
          ++ align 
          ++ text "text \"(\" ++" ++ break 

          ++ (docList (break ++ text "++ text \",\" ++" ++ break)
            (fun (nm,tp) -> pp_te nm tp) () name_type_list)

          ++ break ++ text "++ text \")\"" 
          ++ unalign
      | Constructor("bool",[]) -> tprintf "%b" (text vname)
      | Constructor("int",[]) -> tprintf "%d" (text vname)
      | Constructor("string",[]) -> tprintf "\\\"%s\\\"" (text vname)
      | Constructor(cn,[]) -> dprintf "d_%s () %s" cn vname
      | Constructor("list",[b]) -> 
          tprintf "[@[%a@]]" 
            (text "(docList (chr ';' ++ break)" ++ break ++
             text "(fun elt -> " ++
             (pp_te "elt" b) ++ text "))" ++ break ++ text vname)
      | Constructor("option",[b]) -> 
          tprintf "@[%a@]" 
            (text "(docOpt (fun () elt -> " ++ 
             (pp_te "elt" b) ++ text "))" ++ break ++ text vname)

      | Constructor("node",[b]) -> 
          let newname = gensym () in 
          dprintf "let %s = (%s).node in@! " newname vname 
          ++ pp_te newname b 

      | Constructor(cn,tel) -> 
          dprintf "d_%s %a () %s" cn 
            (docList (break) (fun te -> 
              text "(fun () v ->" 
              ++ break ++ align
              ++ pp_te "v" te 
              ++ unalign ++ text ")"
            )) tel vname 
    end 
    in 
    docList line pp_td () tdl 
  end
  )) () tdll

let visitActionText = "
(** Different visiting actions. *)
type 'a visitAction = 
    SkipChildren                        (** Do not visit the children. Return 
                                            the node as it is. *)
  | DoChildren                          (** Continue with the children of this 
                                            node. Rebuild the node on return 
                                            if any of the children changes 
                                            (use == test) *)
  | ChangeTo of 'a                      (** Replace the expression with the 
                                            given one *)
  | ChangeDoChildrenPost of 'a * ('a -> 'a) (** First consider that the entire 
                                           exp is replaced by the first 
                                           parameter. Then continue with 
                                           the children. On return rebuild 
                                           the node if any of the children 
                                           has changed and then apply the 
                                           function on the node *)"
let doVisitText ="
(* visit all the nodes in a type expression *)
let doVisit (vis: 'b)
            (startvisit: 'a -> 'a visitAction)
            (children: 'b -> 'a -> 'a)
            (node: 'a) : 'a =
  let action = startvisit node in
  match action with
    SkipChildren -> node
  | ChangeTo node' -> node'
  | _ -> (* DoChildren and ChangeDoChildrenPost *)
      let nodepre = match action with
        ChangeDoChildrenPost (node', _) -> node'
      | _ -> node                                                                     in
      let nodepost = children vis nodepre in
      match action with
        ChangeDoChildrenPost (_, f) -> f nodepost
      | _ -> nodepost"

let tmpify_regexp = Str.regexp "\\."

let tmpify s = 
  "tmp_" ^ (Str.global_replace tmpify_regexp "_" s)

(* emit a visitor for a type_def list *)
let emit_v (basename : string) (tdll : type_def list list) = 
  text visitActionText ++ line ++ 

  let tdl = List.flatten tdll in 

  let first = ref true in 
  let rec pv_td (s,tl,ti) = begin 
    (* FIXME TL *) 
    dprintf "%s visit_%s (vis : %sVisitor) v =@!  "
      (if !first then (first := false ; "let rec") else "and") s 
      basename
    ++ align
    ++ dprintf "doVisit vis vis#v_%s visit_children_%s v@!" s s 
    ++ unalign
    ++ dprintf "and visit_children_%s (vis : %sVisitor) v =@!  " s basename
    ++ align
    ++ pv_ti "v" ti

    ++ unalign
  
  end and pv_tel n_t_lst eq_str neq_doc_maker = begin 
    (* e.g. Foo(a,b) -> let temp_a = visit a in
                        let temp_b = visit b in
                        if (temp_a == a && temp_b == b)
                          _
                        else 
                          Foo(temp_a, temp_b) *)
    (docList (nil) (fun (nm,te) ->
      dprintf "let %s = " (tmpify nm)
      ++ align 
      ++ pv_te nm te 
      ++ unalign 
      ++ text " in" 
      ++ line 
    )) () n_t_lst
    ++
    dprintf "if (%a)@?@[then %s@]@?@[else %a@]"
      (docList (break ++ text "&&" ++ break)
        (fun (nm,te) -> dprintf "%s == %s" (tmpify nm) nm)) n_t_lst
      eq_str
      neq_doc_maker n_t_lst
  end and pv_ti (vname : string) (ti : type_info) = begin
    match ti with 
    | OUnion(cdl) -> 
      doc_from_union vname cdl (fun cname n_t_lst -> match n_t_lst with
      | [] -> text vname
      | _ -> pv_tel n_t_lst vname
        (fun () n_t_lst -> dprintf "(%s(%a))" cname
          (docList (break ++ text "," ++ break)
            (fun (nm,te) -> text (tmpify nm))) n_t_lst)
      ) (* end of: doc_from_union *) 
    | ORecord(fdl) ->  
        let ntl = List.map (fun (mut,nm,pte) -> 
          let nm' = Printf.sprintf "%s.%s" vname nm in 
          (nm',pte)) fdl in 
        pv_tel ntl vname
        (fun () n_t_lst -> 
          let triple_list = List.combine n_t_lst fdl in 
          dprintf "{@[ %a @]}" 
          (docList (break ++ break)
            (fun ((nm,te),(mut,fldname,pte)) -> 
              dprintf "%s = %s;" fldname (tmpify nm))
            ) triple_list)
    | OType(te) -> pv_te vname te 
  end and pv_te (vname : string) (te : typexpr) = begin match te with 
    | Tuple(el) -> 
          let en = List.map (fun e -> gensym()) el in 
          dprintf "let (@[%a@]) =@?%s in@!"
            (docList (chr ',' ++ break) (fun s -> text s)) en 
            vname 
          ++ 
          let name_type_list = List.combine en el in 
          pv_tel name_type_list vname 
            (fun n_t_lst -> dprintf "(%a)" 
              (docList (break ++ text "," ++ break)
                (fun (nm,te) -> text (tmpify nm))))
    | Constructor(a,[]) -> 
      begin 
      if List.exists ( fun (n,tl,ti) -> n = a) tdl then 
        (* FIXME TL *) 
        dprintf "visit_%s vis %s" a vname 
      else 
        dprintf "%s (* no %s visitor *)" vname a
      end 
    | Constructor("list",[b]) -> 
      text "List.map (fun v -> " 
      ++ align
      ++ pv_te "v" b 
      ++ unalign 
      ++ dprintf ") %s" vname 
    | Constructor("option",[b]) -> 
      dprintf "match %s with @[None -> %s@]@? | Some(x) -> (" vname vname
      ++ pv_tel [("x",b)] vname 
        (fun n_t_lst -> dprintf "(Some(%a))"
            (fun n_t_lst -> dprintf "(%a)" 
              (docList (break ++ text "," ++ break)
                (fun (nm,te) -> text (tmpify nm)))))
      ++ text ")"
    | Constructor("node",[b]) -> 
      text "{ v with node = let v = v.node in "  
      ++ align 
      ++ pv_te "v" b 
      ++ unalign 
      ++ text "; }" 

    | _ -> text vname
  end in

  dprintf "class type %sVisitor = object@!  " 
    basename ++
  align ++
  (docList nil (fun (s,tl,ti) ->
    dprintf "method v_%s : (%a%s) -> (%a%s) visitAction@!" s 
      doc_from_ticks tl s 
      doc_from_ticks tl s 
  )) () tdl ++
  unalign ++ 
  text "end\n" ++

  dprintf "class nop%sVisitor : %sVisitor = object@!  " 
    (String.capitalize basename) basename
    ++
  align ++
  (docList nil (fun (s,tl,ti) ->
    dprintf "method v_%s : @[%a((%a%s) -> ((%a%s) visitAction))@] =@?@[fun _ ->@?DoChildren@]@!" s 
      (docList nil (fun t -> dprintf "'%s.@?" t)) tl
      doc_from_ticks tl s 
      doc_from_ticks tl s
  )) () tdl ++
  unalign ++ 
  text "end\n" ++

  text doVisitText ++ line ++ 

  (docList line pv_td) () tdl ++ 

  line
