open Pretty
open Testtype

let rec d_myint () (a1 : myint) =
  begin match a1 with
    Int a2 -> text "Int(" ++ break ++  dprintf "%d" a2 ++ text ")"
  end

let rec d_foo () (a3 : foo) =
  begin match a3 with
    This -> text "This"
  | That(a4,a5) -> text "That(" ++ break ++  d_myint () a4 ++ text "," ++
                   d_myint () a5 ++ text ")"
  end

and d_bar () (a6 : bar) =
  text "{ " ++ align ++ text "x = " ++ break ++ d_myint () a6.x ++ text ";"
  ++ line ++
  text "y = " ++ break ++ dprintf "\"%s\"" a6.y ++ text ";" ++ line ++
  text "z = " ++ break ++ d_foo () a6.z ++ unalign ++ text "; }" 

and d_fred () (a7 : fred) =
  dprintf "@[%a@]" (docOpt (fun () elt -> dprintf "[@[%a@]]" (docList (chr ';' ++ break)
                                                             (fun elt -> d_myint () elt))
                                                             elt)) a7

and d_joe () (a8 : joe) =
  let (a9, a10) = a8 in
    text "(" ++ d_myint () a9 ++ text "," ++ d_myint () a10 ++ text ")"

and d_intpair (d_a : unit -> 'a -> doc) () (a11 : 'a intpair) =
  let (a12, a13) = a11 in
    text "(" ++ (d_a () a12) ++ text "," ++ dprintf "%d" a13 ++ text ")"

let rec d_barintpair () (a14 : barintpair) =
  d_intpair (fun () v -> d_bar () v) () a14

and d_stringintpair () (a15 : stringintpair) =
  d_intpair (fun () v -> dprintf "\"%s\"" v) () a15

(** Different visiting actions. *)
type 'a visitAction = 
    SkipChildren                        (** Do not visit the children. Return 
                                            the node as it is. *)
  | DoChildren                          (** Continue with the children of this 
                                            node. Rebuild the node on return 
                                            if any of the children changes 
                                            (use == test) *)
  | ChangeTo of 'a                      (** Replace the expression with the 
                                            given one *)
  | ChangeDoChildrenPost of 'a * ('a -> 'a) (** First consider that the entire 
                                           exp is replaced by the first 
                                           parameter. Then continue with 
                                           the children. On return rebuild 
                                           the node if any of the children 
                                           has changed and then apply the 
                                           function on the node *)
class type testtypeVisitor = object
  method v_myint : (myint) -> (myint) visitAction
  method v_foo : (foo) -> (foo) visitAction
  method v_bar : (bar) -> (bar) visitAction
  method v_fred : (fred) -> (fred) visitAction
  method v_joe : (joe) -> (joe) visitAction
  method v_intpair : ('a intpair) -> ('a intpair) visitAction
  method v_barintpair : (barintpair) -> (barintpair) visitAction
  method v_stringintpair : (stringintpair) -> (stringintpair) visitAction
end
class nopTesttypeVisitor : testtypeVisitor = object
  method v_myint : ((myint) -> ((myint) visitAction)) = fun _ -> DoChildren
  method v_foo : ((foo) -> ((foo) visitAction)) = fun _ -> DoChildren
  method v_bar : ((bar) -> ((bar) visitAction)) = fun _ -> DoChildren
  method v_fred : ((fred) -> ((fred) visitAction)) = fun _ -> DoChildren
  method v_joe : ((joe) -> ((joe) visitAction)) = fun _ -> DoChildren
  method v_intpair : 'a. (('a intpair) -> (('a intpair) visitAction)) =
  fun _ -> DoChildren
  method v_barintpair : ((barintpair) -> ((barintpair) visitAction)) =
  fun _ -> DoChildren
  method v_stringintpair : ((stringintpair) -> ((stringintpair) visitAction)) =
  fun _ -> DoChildren
end

(* visit all the nodes in a type expression *)
let doVisit (vis: 'b)
            (startvisit: 'a -> 'a visitAction)
            (children: 'b -> 'a -> 'a)
            (node: 'a) : 'a =
  let action = startvisit node in
  match action with
    SkipChildren -> node
  | ChangeTo node' -> node'
  | _ -> (* DoChildren and ChangeDoChildrenPost *)
      let nodepre = match action with
        ChangeDoChildrenPost (node', _) -> node'
      | _ -> node                                                                     in
      let nodepost = children vis nodepre in
      match action with
        ChangeDoChildrenPost (_, f) -> f nodepost
      | _ -> nodepost
let rec visit_myint (vis : testtypeVisitor) v =
  doVisit vis vis#v_myint visit_children_myint v
and visit_children_myint (vis : testtypeVisitor) v =
  begin match v with
    Int a16 -> let tmp_a16 = a16 (* no int visitor *) in
               if (tmp_a16 == a16) then v else (Int(tmp_a16))
  end
and visit_foo (vis : testtypeVisitor) v =
  doVisit vis vis#v_foo visit_children_foo v
and visit_children_foo (vis : testtypeVisitor) v =
  begin match v with
    This -> v
  | That(a17,a18) -> let tmp_a17 = visit_myint vis a17 in
                     let tmp_a18 = visit_myint vis a18 in
                     if (tmp_a17 == a17 && tmp_a18 == a18) then v
                     else (That(tmp_a17 , tmp_a18))
  end
and visit_bar (vis : testtypeVisitor) v =
  doVisit vis vis#v_bar visit_children_bar v
and visit_children_bar (vis : testtypeVisitor) v =
  let tmp_v_x = visit_myint vis v.x in
  let tmp_v_y = v.y (* no string visitor *) in
  let tmp_v_z = visit_foo vis v.z in
  if (tmp_v_x == v.x && tmp_v_y == v.y && tmp_v_z == v.z) then v
  else { x = tmp_v_x;  y = tmp_v_y;  z = tmp_v_z; }
and visit_fred (vis : testtypeVisitor) v =
  doVisit vis vis#v_fred visit_children_fred v
and visit_children_fred (vis : testtypeVisitor) v =
  match v with None -> v
   | Some(x) -> (let tmp_x = List.map (fun v -> visit_myint vis v) x in
  if (tmp_x == x) then v else (Some((tmp_x))))
and visit_joe (vis : testtypeVisitor) v =
  doVisit vis vis#v_joe visit_children_joe v
and visit_children_joe (vis : testtypeVisitor) v =
  let (a19, a20) = v in
  let tmp_a19 = visit_myint vis a19 in
  let tmp_a20 = visit_myint vis a20 in
  if (tmp_a19 == a19 && tmp_a20 == a20) then v else (tmp_a19 , tmp_a20)
and visit_intpair (vis : testtypeVisitor) v =
  doVisit vis vis#v_intpair visit_children_intpair v
and visit_children_intpair (vis : testtypeVisitor) v =
  let (a21, a22) = v in
  let tmp_a21 = a21 in
  let tmp_a22 = a22 (* no int visitor *) in
  if (tmp_a21 == a21 && tmp_a22 == a22) then v else (tmp_a21 , tmp_a22)
and visit_barintpair (vis : testtypeVisitor) v =
  doVisit vis vis#v_barintpair visit_children_barintpair v
and visit_children_barintpair (vis : testtypeVisitor) v =
  v
and visit_stringintpair (vis : testtypeVisitor) v =
  doVisit vis vis#v_stringintpair visit_children_stringintpair v
and visit_children_stringintpair (vis : testtypeVisitor) v =
  v
